/*
 * lockfile.c  -  File locking
 *
 * Copyright (C) 2003-2006 Gero Kuhlmann <gero@gkminix.han.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: lockfile.c,v 1.6 2005/11/13 20:14:34 gkminix Exp $
 */

#include <common.h>
#include <nblib.h>
#include "privlib.h"


/*
 * Check for flock() system call
 */
#ifdef HAVE_FLOCK
# ifdef HAVE_SYS_FILE_H
#  include <sys/file.h>
# else
#  undef HAVE_FLOCK
# endif
#endif



/*
 * Length of a PID string in a lock file
 */
#define PIDLEN	16



/*
 * Variables local to this module
 */
static int isinit = FALSE;			/* initialization flag */
static int locknum = 0;				/* number of lock files */



/*
 * Structure holding all current lockfiles
 */
struct lockstruct {
	char              *fname;
	struct lockstruct *next;
};

static struct lockstruct *locklist = NULL;




/*
 * Create a new lock file structure
 */
static struct lockstruct *newlock(fname)
char *fname;
{
  struct lockstruct *lp;

  lp = (struct lockstruct *)nbmalloc(sizeof(struct lockstruct));
  copystr(&(lp->fname), fname);
  lp->next = locklist;
  locklist = lp;
  locknum++;
  return(lp);
}



/*
 * Find lock file in list
 */
static struct lockstruct *findlock(fname)
char *fname;
{
  struct lockstruct *lp;

  lp = locklist;
  while (lp != NULL) {
	if (!strcmp(lp->fname, fname))
		break;
	lp = lp->next;
  }
  return(lp);
}



/*
 * Remove a lock file
 */
static void removelock(curlock)
struct lockstruct *curlock;
{
  struct lockstruct *lp;

  /* Unlink the file */
  (void)unlink(curlock->fname);

  /* Remove the lock file from the list */
  if (curlock == locklist)
	locklist = curlock->next;
  else {
	lp = locklist;
	while (lp != NULL && lp->next != curlock)
		lp = lp->next;
	if (lp->next == curlock)
		lp->next = curlock->next;
  }
  locknum--;

  /* Delete the lock file entry */
  free(curlock->fname);
  free(curlock);
}



/*
 * Function to be called upon program termination
 */
static void unlockall()
{
  while (locklist != NULL)
	removelock(locklist);
}



/*
 * Initialize module
 */
static inline void initmodule()
{
  int ret;

  if (!isinit) {
	if ((ret = nbatexit(&unlockall)) != 0) {
		prnerr("unable to set lock file exit function");
		exit(EXIT_INTERNAL);
	}

	locknum = 0;
	locklist = NULL;
	isinit = TRUE;
  }
}



/*
 * Write our process PID into the lock file
 */
static void writepid(fd, curlock)
FILE *fd;
struct lockstruct *curlock;
{
  char pidbuf[PIDLEN + 1];
  int i;
  pid_t pid;

  pid = getpid();
  for (i = PIDLEN - 1; i >= 0; i--) {
	pidbuf[i] = (char)((int)(pid % 10) + '0');
	pid /= 10;
  }
  pidbuf[PIDLEN] = '\0';
  if (fprintf(fd, "%s", pidbuf) != PIDLEN) {
	prnerr("error writing PID into file %s", curlock->fname);
	exit(EXIT_LOCK);
  }
}



/*
 * Read a process PID out of a lock file
 */
static pid_t readpid(fd, curlock)
FILE *fd;
struct lockstruct *curlock;
{
  pid_t pid;
  int c;

  /* Check if we are at EOF already */
  if (feof(fd))
	return((pid_t)0);

  /* Read a PID out of the file */
  pid = (pid_t)0;
  while ((c = fgetc(fd)) != EOF) {
	if (c == '\n')
		break;
	if (!isdigit(c))
		return((pid_t)0);
	pid *= 10;
	pid += c - '0';
  }
  if (c == EOF && !feof(fd)) {
	prnerr("error reading PID from file %s", curlock->fname);
	exit(EXIT_LOCK);
  }
  return(pid);
}



/*
 * Lock the lock file
 */
static void locklock(fd, curlock)
FILE *fd;
struct lockstruct *curlock;
{
  /* Only lock the lock file if we have flock() */
#ifdef HAVE_FLOCK
  if (flock(fileno(fd), LOCK_EX | LOCK_NB) == -1) {
	prnerr("unable to lock file %s", curlock->fname);
	exit(EXIT_LOCK);
  }
#endif
}



/*
 * Lock a file name
 */
int filelock(lockfile)
char *lockfile;
{
  FILE *fd;
  struct lockstruct *lp;
  int ret = TRUE;

  /* Initialize the module */
  initmodule();

  /* Check that we have a lock file name */
  if (lockfile == NULL)
	return(TRUE);

  /* If we can open the lock file for reading, the file is locked */
  if ((fd = fopen(lockfile, "r")) != NULL) {
	(void)fclose(fd);
	return(FALSE);
  }

  /* Otherwise create a new lock file */
  if ((fd = fopen(lockfile, "w")) == NULL) {
	prnerr("error creating file %s", lockfile);
	exit(EXIT_LOCK);
  }
  lp = newlock(lockfile);
  locklock(fd, lp);
  writepid(fd, lp);

  /* Close the lock file again */
  if (fclose(fd) != 0) {
	prnerr("error closing file %s", lockfile);
	exit(EXIT_LOCK);
  }
  return(ret);
}



/*
 * Unlock a file name
 */
void fileunlock(lockfile)
char *lockfile;
{
  FILE *fd;
  struct lockstruct *lp;

  /* Check that we have a lock file name */
  if (lockfile == NULL || (lp = findlock(lockfile)) == NULL)
	return;

  /* The lock file has to exist */
  if ((fd = fopen(lp->fname, "r")) == NULL) {
	prnerr("error opening file %s", lockfile);
	exit(EXIT_LOCK);
  }
  locklock(fd, lp);

  /* Read the PID from lock file */
  if (readpid(fd, lp) != getpid()) {
	(void)fclose(fd);
	prnerr("invalid PID in file %s", lockfile);
	exit(EXIT_LOCK);
  }

  /* Delete the lock file */
  removelock(lp);

  /* Finally close the lock file */
  if (fclose(fd) != 0) {
	prnerr("error closing file %s", lockfile);
	exit(EXIT_LOCK);
  }
}



/*
 * Check for a valid lock file. It returns FALSE if the lock file is not valid.
 */
int checklock(lockfile)
char *lockfile;
{
  FILE *fd;
  struct lockstruct *lp;
  int ret;

  /* Check that we have a lock file name */
  if (lockfile == NULL || (lp = findlock(lockfile)) == NULL)
	return(TRUE);

  /* Check if lock file exists */
  if ((fd = fopen(lp->fname, "r")) == NULL)
	return(FALSE);

  /* Read the PID from lock file */
  ret = (readpid(fd, lp) == getpid());

  /* Finally close the lock file */
  if (fclose(fd) != 0) {
	prnerr("error closing file %s", lockfile);
	exit(EXIT_LOCK);
  }
  return(ret);
}

